import { Endpoints } from "../utils/Endpoints";
import { isAxiosError } from "../utils/Methods";
import { If } from "../utils/Util";
import { BaseClient, BaseClientConfig } from "./BaseClient";
import {
	RESTAPIAccountSuccess,
	RESTAPIOauth2LoginSuccess,
} from "@iFunny/ifunny-api-types";
import { ErrorCodes } from "../errors/ErrorCodes";
import iFunnyError from "../errors/iFunnyError";
import { AxiosResponse } from "axios";
import { APIClientUser as ClientPayload } from "@iFunny/ifunny-api-types";
import { RESTAPIErrorResponse } from "@ifunny/ifunny-api-types";

interface ClientConfig extends Partial<BaseClientConfig> {}

/**
 * Client for the Discord API.
 * @extends BaseClient
 */
export class Client<Authorized extends boolean = boolean> extends BaseClient {
	constructor(config: ClientConfig, payload: Partial<ClientPayload>) {
		super(config, payload);
	}

	/**
	 * The Basic Auth token for the client\
	 * If one does not already exist, it will be created
	 */
	get basic(): string {
		return (this._basic ||= this.generateBasicToken());
	}

	/**
	 * The bearer token for the client. If the client has not logged in, this will be null
	 */
	get bearer(): If<Authorized, string, null> {
		return (this._bearer ||= null) as If<Authorized, string, null>;
	}

	set bearer(value: string | null) {
		typeof value === "string" ? (this._bearer = value) : null;
		this.instance.defaults.headers.common.Authorization = this.authorization;
	}

	/**
	 * Authorization string to use in request headers
	 */
	get authorization(): string {
		return this.bearer ? `Bearer ${this.bearer}` : `Basic ${this.basic}`;
	}

	/**
	 * Is the client using a bearer token to make requests?
	 */
	isAuthorized(): this is Client<true> {
		return !!this.bearer;
	}

	async fetch(): Promise<this> {
		if (!this.isAuthorized()) {
			throw new iFunnyError(ErrorCodes.ClientNotLoggedIn);
		}
		try {
			const response = await this.instance.get<RESTAPIAccountSuccess>(
				Endpoints.account
			);
			let data = response.data;
			this.payload = data.data;
			return this;
		} catch (error) {
			if (error instanceof Error && isAxiosError(error)) {
				const resp = error.response as AxiosResponse<RESTAPIErrorResponse>;
				switch (error?.response?.status) {
					case 401:
						throw new iFunnyError(
							ErrorCodes.InvalidGrant,
							resp.data.error_description
						);
				}
			}
			throw new iFunnyError(ErrorCodes.UnknownError, error);
		}
	}

	/**
	 *
	 * @param email The email to login with
	 * @param password The password to login with
	 * @returns The instance of the client
	 */
	async login(email: string, password: string): Promise<this> {
		if (!this.isAuthorized()) {
			return await this.fetch();
		}
		let response = await this.instance.post<RESTAPIOauth2LoginSuccess>(
			Endpoints.token,
			{
				grant_type: "password",
				username: email,
				password,
			}
		);
		this.bearer = response.data.access_token;
		return this;
	}
}
