const kCode = Symbol("code");
import { ErrorCodes } from "./ErrorCodes";
import { Messages } from "./Messages";

/**
 * A class for representing an error thrown by iFunny.ts
 */
export class iFunnytsError extends Error {
	/**
	 * The error code for the error
	 */
	[kCode]: ErrorCodes;
	/**
	 * @param code The error code for the error
	 * @param args Args to pass into the message formatter function
	 */
	constructor(code: ErrorCodes, ...args: any[]) {
		super(message(code, args));
		this[kCode] = code;
		Error.captureStackTrace?.(this, iFunnytsError);
	}

	/**
	 * The error code for the error
	 */
	get code() {
		return this[kCode];
	}

	/**
	 * The name of the error
	 */
	get name() {
		return `${ErrorCodes[this.code]} [${this.code}]`;
	}
}

/**
 * Format the message for an error.
 * @param code The error code
 * @param args Arguments to pass for util format or as function args
 * @returns Formatted string
 */
function message<K extends keyof typeof Messages>(code: K, ...args: any[]): string {
	if (typeof code !== "number")
		throw new Error("Error code must be a valid iFunnytsErrorCode");
	const msg = Messages[code];
	if (!msg) throw new Error(`An invalid error code was used: ${code}.`);
	// @ts-ignore
	if (typeof msg === "function") return msg(...args);
	if (!args?.length) return msg;
	args.unshift(msg);
	return String(...args);
}

export default iFunnytsError;
