import { Client } from "../client/Client";
import { JSONObject } from "../utils/Util";

const DefaultObjectEndpoint = "/account";

/**
 * Base payload that all other payloads will inherit
 */
export interface BasePayload extends JSONObject {
	id: string;
}

/**
 * Base class for all structures.
 */
export class Base<Payload extends BasePayload = BasePayload> {
	/**
	 * Client instance attached to the Object
	 */
	public readonly client: Client;
	/**
	 * The payload of the object.
	 */
	public payload: Partial<Payload>;
	/**
	 * Endpoint url the object will request to
	 */
	private _endpoint_url: string = "";

	/**
	 * @param client The client instance
	 * @param payload The payload of the object
	 */
	constructor(client: Client, payload: Partial<Payload> = {}) {
		this.client = client;
		this.payload = payload ?? {};
	}

	protected get endpoint_url(): string {
		return this._endpoint_url;
	}

	protected set endpoint_url(value: string) {
		this._endpoint_url = `${value}`;
	}

	get id(): string {
		return this.get("id") as string;
	}

	/**
	 * Base method for getting a value from the payload.
	 * @param key The payload key to look up
	 * @returns The payload value or null if it doesn't exist
	 */
	get<K extends keyof Payload>(key: K): NonNullable<Payload[K]> | null {
		return this.payload[key] ?? null;
	}

	/**
	 * Base method for checking if a value exists in the payload.
	 * @param key The payload key to look up
	 * @returns Whether the payload value exists
	 */
	has<K extends keyof Payload>(key: K): boolean {
		return this.payload[key] !== undefined;
	}

	/**
	 * Creates a new instance of the same structure with the same payload.
	 * @returns A clone of the current instance
	 */
	protected _clone(): typeof this {
		return Object.assign(Object.create(this), this);
	}

	/**
	 * Updates the payload of the object.
	 * @param payload The payload to merge into the current instance
	 * @returns The current instance
	 */
	protected _update_payload(payload: Partial<Payload>): this {
		this.payload = Object.assign(this.payload, payload);
		return this;
	}

	valueOf() {
		return this.id;
	}
}
